// ****************************************************************
//
//   IBM Confidential
//   OCO Source Materials
//   <<com.ibm.etools.cobol>>
//   (C) Copyright IBM Corp. 2001,2004 - All Rights Reserved.
//   US Government Users Restricted Rights - Use, duplication or disclosure
//   restricted by GSA ADP Schedule Contract with IBM Corp.
//
// ****************************************************************

package com.ibm.etools.cobol.importer;

import java.util.*;


import com.ibm.ivj.eab.record.cobol.CobolType;
import com.ibm.etools.cobol.COBOLUsageValues;

/**
 * This class provides the logic to determine whether an initial value
 * is valid for a particular CobolType object.
 */

public class CobolTypeUtilities {
	private static final String copyright = "(c) Copyright IBM Corporation 2001,2004.";

 	
/**
 * This method returns true if the String s contains at least one occurrence
 * of the character c.
 *
 * @return boolean
 * @param s java.lang.String
 * @param c char
 */
public static boolean contains(String s, char c) {

	if (s.indexOf(c) != -1)
		return true;
	
	return false;
}
/**
 * This method expands a PIC string into its full form.
 * eg. A(4)V9(2) becomes AAAAV99
 *
 * @return java.lang.String
 * @param pic java.lang.String
 */
public static String expandPic(String pic) {

	StringBuffer ePic = new StringBuffer();
	int picLength = pic.length();
	int i, j, index;
	int repetitions;
	char charToRepeat, nextChar;
	
	for (i=0; i < picLength; i++) {
		nextChar = pic.charAt(i);
		
		if (nextChar == '(') {
			
			charToRepeat = pic.charAt(i-1);
			
			// look for next ')'
			index = pic.indexOf(')', i);
			if (index == -1)
				return pic;

			// get the number of repetitions
			repetitions = (new Integer(pic.substring(i+1, index))).intValue();

			i = index;

			// we have already written one of the characters
			repetitions--;  

			// expand these brackets
			while (repetitions > 0) {
				ePic.append(charToRepeat);
				repetitions--;
			}
		}

		else 
			ePic.append(nextChar);
	}
	
	return ePic.toString();
}

/**
 *  Return the field width based on the picture string
 *  and the field type
 */
public static int getNumericFieldWidth(String sPicString, String sAugString, int nUsage) {

int nWidth = 0;

	int numberOfNines = getNumberOfNines(sPicString);
	
	/*
	 * Use the usage as the switch
	 * the augmentor is used only to determine if there is a separate sign
	 */
	switch(nUsage) {
		case COBOLUsageValues.FLOAT:
			nWidth = 4;
			break;

		case COBOLUsageValues.DOUBLE:
			nWidth = 8;		
			break;

		case COBOLUsageValues.BINARY:
			if(numberOfNines > 9) {
				nWidth = 8;
			}
			else if(numberOfNines > 4) {
				nWidth = 4;
			}
			else {
				nWidth = 2;
			}	
			break;

		case COBOLUsageValues.PACKED_DECIMAL:
			nWidth = numberOfNines/2 + 1;		
			break;
		
		case COBOLUsageValues.DISPLAY:
			/*
			 * If the sign is leading or trailing separate,
			 * add one space to the number of nines
			 */
			if(sAugString.indexOf("SEPARATE") != -1) {
				nWidth = numberOfNines + 1;
			}
			else {
				nWidth = numberOfNines;
			}
		
			break;
		
		default:
			nWidth = numberOfNines;
	} /* END switch usage */	 
	
	return nWidth;
}

/**
 * Count the number of 9's in the picture string
 */
public static int getNumberOfNines(String sPicString) {

int numberOfNines = 0;

	for(int i=0; i < sPicString.length(); i++) {
		if(sPicString.charAt(i) == '9') {
			numberOfNines++;
		}
	}
	
	return numberOfNines;
}

/**
 * Determine the virtual decimal point location
 */
public static int getVirtualDecimalPointLocation(String sPicString) {

int virtualDecimalLocation = 0;

	/*
	 * If there is no "P" in the picture string, use the "V"
	 * to determine the decimal point location
	 */
	if(sPicString.indexOf('P') == -1) {
		/*
		 * Specify the location of the virtual decimal point if present
		 */
		int virtualDecimalIndex = sPicString.indexOf('V');
	
		if(virtualDecimalIndex == -1) {
			virtualDecimalLocation = 0;
		}
		else {
			virtualDecimalLocation = sPicString.length() - virtualDecimalIndex - 1;
		}
	}
	/*
	 * If there is a "P" in the picture string,
	 * ignore the "V," since the "P" determines the
	 * decimal point location
	 */
	else {
		/*
		 * Count the number of P's and 9's,
		 * and determine which symbol comes first
		 */
		char firstCharacter = ' ';
		int numberOfNines   = 0;
		int numberOfPs      = 0;
				 
		for(int i=0; i < sPicString.length(); i++) {
			if(sPicString.charAt(i) == '9') {
				if(firstCharacter == ' ') {
					firstCharacter = '9';
				}
				numberOfNines++;
			}
			else if(sPicString.charAt(i) == 'P') {
				if(firstCharacter == ' ') {
					firstCharacter = 'P';
				}
				numberOfPs++;
			}
		}
				 
		/*
		 * If the P's are before the 9's, add them to the location.
		 * If the P's are after the 9's, subtract them
		 */
		if(firstCharacter == 'P') {
			virtualDecimalLocation = numberOfNines + numberOfPs;
		}
		else {
			virtualDecimalLocation = numberOfPs * -1;
		}
	} /* END of if no P in picture string */

	return virtualDecimalLocation;
}

/**
 * This method gets for a given type the alignment boundary
 * required to pack a synchronized field
 */
public static int getAlignmentFactor(CobolType cobolType) {

String sPicString = CobolTypeUtilities.expandPic(cobolType.getPic());
String sAugString = cobolType.getAug();
int alignmentFactor;

	/*
	 * The alignment factor is based on the type
	 */
	if(sAugString.indexOf("COMP1") != -1) {
		alignmentFactor = 4;
	}
	else if(sAugString.indexOf("COMP2") != -1) {
		alignmentFactor = 8;
	}
	else if(sAugString.indexOf("COMP3") != -1) {
		alignmentFactor = 1;
	}
	else if((sAugString.indexOf("COMP4") != -1) ||
		(sAugString.indexOf("COMP5") != -1) || 
		(sAugString.indexOf("COMP") != -1) ||
		(sAugString.indexOf("BINARY") != -1)) {
		/*
		 * The alignment factor is based on the size of the binary field
		 */
		int numberOfNines = CobolTypeUtilities.getNumberOfNines(sPicString);
		if(numberOfNines >=  5) {
			alignmentFactor = 4;
		}	
		else {
		 	alignmentFactor = 2;
		}
	}
	/*
	 * The type does not require alignment
	 */
	else {
		alignmentFactor = 1;		
	}
	
	return alignmentFactor;		
}

/**
 * This method gets the additional bytes
 * required to pack a synchronized field
 */
public static int getPackingBytes(String positionString, int alignmentFactor) {

int positionNumber;
int positionRemainder;
int packingBytes;
	
	/*
	 * Divide the current position by the alignment factor and take the remainder
	 * The position must be incremented by the factor minus the remainder in order
	 * to place the next offset on that byte boundary
	 */
	try {
		positionNumber = Integer.valueOf(positionString).intValue();
	}
	catch(Exception ex) {
		/*
		 * The offset has a formula for a variable length array.
		 * Since the rows will line up on the same boundaries,
		 * take the first row
		 */
		positionNumber = evaluateFormula(positionString);
	}	 
	
	positionRemainder = positionNumber % alignmentFactor;
	if(positionRemainder > 0) {
		packingBytes = alignmentFactor - positionRemainder;
	}
	else {
		packingBytes = 0;
	}
	
	return packingBytes;		
}

/**
 * This method creates the offset formula.
 * For records that do not include an OCCURS DEPENDING ON clause,
 * the offset is taken from the field's size.
 * After an ODO is declared, the offset must be expressed as a formula
 * in which the value of the depending on field determines the offset at runtime
 */
public static String createOffset(String parentOffsetFormula, int parentOffsetNumber, String relativeOffsetFormula, int relativeOffsetNumber) {

String stringOffset;
int formulaOffset;
int totalOffsetNumber = parentOffsetNumber + relativeOffsetNumber;

	/*
	 * Combine the two formula
	 */
	String totalOffsetFormula = addOffsetFormulas(parentOffsetFormula, relativeOffsetFormula);
		
	/*
	 * The final offset is the parentOffsetFormula if present plus the numeric offset
	 */
	try {	 
		if(totalOffsetFormula.length() == 0) {
			formulaOffset = 0;
		}
		else {
			formulaOffset = Integer.valueOf(totalOffsetFormula).intValue();
		}
		stringOffset = (new Integer(formulaOffset + totalOffsetNumber)).toString();
	}
	catch(Exception ex) {
		if(totalOffsetNumber > 0) {
			stringOffset = "add(" + totalOffsetFormula + "," + totalOffsetNumber + ")";
		}
		else {
			stringOffset = totalOffsetFormula;
		}
	}
			
	return stringOffset;	
}

/**
 * This method adds the relative offset to the current offset
 * If the offsets are numbers, simple addition is required.
 * If the offsets are formulas, a new formula must be created
 */
public static String addOffsetFormulas(String currentOffsetFormula, String relativeOffsetFormula) {

String stringFormula;
int relativeOffsetNumber;
int currentOffsetNumber;

	/*
	 * If the relativeOffset is a number, add it to the currentOffsetFormula
	 */
	try {
		if((currentOffsetFormula.length() == 0) ||
			(currentOffsetFormula.compareTo("0") == 0)) {
			return relativeOffsetFormula;
		}
		else if((relativeOffsetFormula.length() == 0) ||
			(relativeOffsetFormula.compareTo("0") ==0)) {
			return currentOffsetFormula;
		}

		currentOffsetNumber  = Integer.valueOf(currentOffsetFormula).intValue();	
		relativeOffsetNumber = Integer.valueOf(relativeOffsetFormula).intValue();
				
		stringFormula = (new Integer(currentOffsetNumber + relativeOffsetNumber)).toString();
	}
	catch(Exception ex) {
		stringFormula = "add(" + currentOffsetFormula + "," + relativeOffsetFormula + ")";
	}	
		 			
	return stringFormula;	
}

/**
 * This method multiplies the relative offset to the current offset
 * If the offsets are numbers, simple multiplication is required.
 * If the offsets are formulas, a new formula must be created
 */
public static String multiplyOffsetFormulas(String strideFormula, String dimensionString) {

String stringFormula;
int strideNumber;
int dimensionNumber;
	
	/*
	 * If the stride formula is a number, multiply is by the dimension to get the size
	 */
	try {
		if(strideFormula.length() == 0) {
			strideNumber = 0;
		}
		else {
			strideNumber  = Integer.valueOf(strideFormula).intValue();	
		}
		dimensionNumber = Integer.valueOf(dimensionString).intValue();

		stringFormula = (new Integer(strideNumber * dimensionNumber)).toString();
	}
	catch(Exception ex) {
		stringFormula = "mpy(" + strideFormula + "," + dimensionString + ")";
	}	
		 			
	return stringFormula;	
}

/**
 * This method creates the size.
 * For records that do not include an OCCURS DEPENDING ON clause,
 * the size is taken from the field's size.
 * After an ODO is declared, the size must be expressed as a formula
 * in which the value of the depending on field determines the offset at runtime
 */
public static String createSize(String relativeOffsetFormula, int relativeOffsetNumber) {

	try {
		if(relativeOffsetFormula.length() == 0) {
			return (new Integer(relativeOffsetNumber)).toString();
		}
		else if(relativeOffsetNumber == 0) {
			return relativeOffsetFormula;
		}
		else {
			int formulaNumber = Integer.valueOf(relativeOffsetFormula).intValue();
			return (new Integer(formulaNumber + relativeOffsetNumber)).toString();
		}
	}
	catch(Exception ex) {
		return "add(" + relativeOffsetFormula + "," + relativeOffsetNumber + ")";
	}	
}

/**
 * This method creates an expression that specifies
 * the ODO control field relative to the top record
 * e.g., "val(1, 2, 3)" means that the control field's
 * location is as follows:
 * "1" - the control field is in the same level-1 record
 * "2" - the control field is in the second sub-element of the level-1 record
 * "3" - the control field is the third sub-element of the level above
 */
public static String createODOValExpression(Vector fieldPositionVector, int relativePosition) {

String valExpression = "val(1";

	for(int i=0; i < fieldPositionVector.size(); i++) {
		if(fieldPositionVector.elementAt(i) != null) {
			valExpression += "," + ((Integer)(fieldPositionVector.elementAt(i))).toString();
		}
	}

	/*
	 * Append this level's position
	 */
	valExpression += "," + relativePosition;
		
	/*
	 * Close the val expression parentheses
	 */
	valExpression += ")";
		
	return valExpression;
}

/**
 * Returns the integer value of the formula string for an offset or size.
 * @return The integer value of the formula string for an offset or size.
 * @param formula The formula to be evaluated.
 * @param valFieldNameMap A map of val() formulas to field names.
 * @param fh The format handler associated with the valFieldNameMap.
 * @throws IllegalArgumentException if the formula is null.
 *
 * The formula can be comprised of the following functions:
 * neg(x)   := -x			// prefix negate
 * add(x,y) := x+y		// infix add
 * sub(x,y) := x-y		// infix subtract
 * mpy(x,y) := x*y		// infix multiply
 * div(x,y) := x/y		// infix divide
 * max(x,y) := max(x,y)
 * min(x,y) := min(x,y)
 *
 * mod(x,y) := x mod y
 *
 * The mod function is defined as mod(x,y) = r where r is the smallest non-negative integer
 * such that x-r is evenly divisible by y. So mod(7,4) is 3, but mod(-7,4) is 1. If y is a
 * power of 2, then mod(x,y) is equal to the bitwise-and of x and y-1.
 *
 * val(1, m, n, o,..)
 *
 * The val function returns the value of a field in the model. The val function takes one
 * or more arguments, and the first argument refers to a level-1 field in the type model and must be either:
 * 	- the name of a level-1 field described in the language model
 * 	- the integer 1 (indicating that the level-1 parent of the current structure is meant)
 * If the first argument to the val function is the integer 1, then and only then are subsequent arguments
 * permitted. These subsequent arguments are integers that the specify the ordinal number within its
 * substructure of the subfield that should be dereferenced.
 */
private static int evaluateFormula (String formula)
{
	int result = 0;

	int index = formula.indexOf("(");

	if (index == -1)				// It's a number not an expression
		return (Integer.parseInt(formula));

	// Determine the outermost function
	String function = formula.substring (0, index);

	if (function.equalsIgnoreCase ("val"))
	{
		result = 1;		
	}
	else if (function.equalsIgnoreCase ("neg"))
	{
		// The new formula is the content between the brackets
		formula = formula.substring (index+1, formula.length()-1);
		result = -1 * evaluateFormula (formula);
	}
	else
	{
		// Get the contents between the outermost brackets
		formula = formula.substring (index+1, formula.length()-1);
		char[] formulaChars = formula.toCharArray();

		// Get the left side and the right side of the operation

		int brackets = 0;
		int i = 0;

		for ( ; i < formulaChars.length; i++)
		{
			if (formulaChars[i] == '(')
				brackets++;
			else if (formulaChars[i] == ')')
				brackets--;
			else if (formulaChars[i] == ',')
			{
				if (brackets == 0)
					break;
			}
		}

		String leftSide = "0";
		String rightSide = "0";

		leftSide = formula.substring (0, i);
		rightSide = formula.substring (i+1);

		if (function.equalsIgnoreCase ("add"))
			result = evaluateFormula (leftSide) + evaluateFormula (rightSide);
		else if (function.equalsIgnoreCase ("mpy"))
			result = evaluateFormula (leftSide) * evaluateFormula (rightSide);
		else if (function.equalsIgnoreCase ("sub"))
			result = evaluateFormula (leftSide) - evaluateFormula (rightSide);
		else if (function.equalsIgnoreCase ("div"))
			result = evaluateFormula (leftSide) / evaluateFormula (rightSide);
		else if (function.equalsIgnoreCase ("max"))
			result = Math.max (evaluateFormula (leftSide), evaluateFormula (rightSide));
		else if (function.equalsIgnoreCase ("min"))
			result = Math.min (evaluateFormula (leftSide),  evaluateFormula (rightSide));
		else if (function.equalsIgnoreCase ("mod"))
			result = evaluateFormula (leftSide) % evaluateFormula (rightSide);
	}

	return (result);
}

/**
 * This method returns true if the given picture string represents
 * an alphabetic item, and false otherwise.  For alphabetic items,
 * the picture string can only contain the symbol 'A'.
 * 
 * @return boolean
 * @param pic java.lang.String
 */
 
public static boolean isAlphabetic(String pic) {
	
	char c;
	int picLength = pic.length();
	
	for (int i = 0; i < picLength; i++) {
		c = pic.charAt(i);
		if (c != 'A')
			return false;
	}

	return true;
}
/**
 * This method returns true if the given picture string represents
 * an alphanumeric item, and false otherwise.  For alphanumeric items,
 * the picture string must consist of either of the following:
 * - the symbol 'X'
 * - combinations of the symbols 'A', 'X', and '9' (a picture string containing
 * all As or all 9s does not define an alphanumeric item)
 * 
 * @return boolean
 * @param pic java.lang.String
 */
 
public static boolean isAlphaNumeric(String pic) {
	char c;
	int picLength = pic.length();
	
	for (int i = 0; i < picLength; i++) {
		c = pic.charAt(i);
		if (!(c == 'A' || c == '9' || c == 'X'))
			return false;
	}

	return true;
}
/**
 * This method returns true if the given picture string represents
 * an alphanumeric-edited item, and false otherwise.  For alphanumeric-edited
 * items, the picture string can contain the following symbols:
 * 	'A', 'X', '9', 'B', '0', '/'
 * - the string must contain at least one 'A' or 'X', and at least one 'B' or
 * '0' (zero) or '/'
 * 
 * @return boolean
 * @param pic java.lang.String
 */

public static boolean isAlphaNumericEdited(String pic) {

	char c;
	int picLength = pic.length();
	
	for (int i = 0; i < picLength; i++) {
		c = pic.charAt(i);
		if (!(c == 'A' || c == '9' || c == 'X' || c == 'B' || c == '0' || c == '/'))
			return false;
	}

	// alphanumeric strings must contain at least one A or X, and at least one B or 0 or /
	return (contains(pic, 'A') || contains(pic, 'X'))
			&& (contains(pic, 'B') || contains(pic, '0') || contains (pic, '/'));
}
/**
 * This method returns true if the given picture string represents
 * a DBCS item, and false otherwise.  For DBCS items, the picture
 * string can only contain the following symbols:
 * 'G' 'B' and 'N'
 * 
 * @return boolean
 * @param pic java.lang.String
 */

public static boolean isDBCS(String pic) {
	char c;
	int picLength = pic.length();
	
	for (int i = 0; i < picLength; i++) {
		c = pic.charAt(i);
		if (!(c == 'G' || c == 'B' || c == 'N'))
			return false;
	}

	return true;
}
/**
 * This method returns true if the picture string represents an
 * external floating point item.  Otherwise, it returns false.
 * A picture string for an external floating point item can contain 
 * the following characters:
 * '9', '+', '-', 'E', 'V', '.'
 * The picture string must contain the exponent character 'E'.
 * 
 * @return boolean
 * @param pic java.lang.String
 */
 
public static boolean isExternalFloatingPoint(String pic) {

	char c;
	boolean externalFloatingPoint = true;

	int picLength = pic.length();

	for (int i=0; i < picLength; i++) {

		c = pic.charAt(i);
			
		if (!(c == '9' || c == '+' || c == '-' || c == 'E' || c == 'V' || c == '.'))
			externalFloatingPoint = false;
	}
	
	if (externalFloatingPoint)
		externalFloatingPoint = contains(pic, 'E');

	return externalFloatingPoint;
}
/**
 * Returrns <code>true</code> if this CobolType is numeric, otherwise <code>false</code>.
 * @return Boolean value indicating if the supplied PIC String matches to a COBOL numeric type.
 * @param picStr A COBOL PIC string.
 */
public static boolean isNumeric(String picString) {

	char c;
	
	// returns true if the pic string contains only the 'S', 'V', '9', and 'P' characters
	for (int i=0; i < picString.length(); i++) {
		c = picString.charAt(i);
		
		if (c == '(') {
			
			for (i = i + 1; i < picString.length(); i++) {
				c = picString.charAt(i);
				
				if (c == ')')
					break;
				// else if c is not numeric, then return false
				else if (!Character.isDigit(c))
					return false;
			}
				
		}
			
		else if (c != 'S' && c != 'V' && c != '9' && c != 'P')
			return false;
	}
	
	return true;
}
/**
 * This method returns true if the given picture string represents
 * a numeric-edited item, and false otherwise.  For numeric-edited items,
 * the picture string can contain the following symbols:
 * 'B', 'P', 'V', 'Z', '9', '0', '/', ',', '.', '+', '-', 'CR', 'DB', '*', '$'
 * 
 * @return boolean
 * @param pic java.lang.String
 */

public static boolean isNumericEdited(String pic) {
	char c;
	int picLength = pic.length();
	
	for (int i = 0; i < picLength; i++) {
		c = pic.charAt(i);
		if (!(c == 'B' || c == 'P' || c == 'V' || c == 'Z' || c == '9' || c == '0'	|| c == '/' || c == ',' 
			|| c == '.' || c == '+' || c == '-' || c == '*' || c == '$' || c == 'C' || c == 'R' || c == 'D'))
				return false;
	}

	return true;
}

public static void validateCodepage(String codepage) throws Exception {
	/*
	 * Test the codepage by trying to convert a byte array
	 * to a String using the specifed codepage
	 * If it fails, the codepage is invalid
	 */
	try {	 	
		byte byteArray[] = {12, 34};
		String s = new String(byteArray, codepage);
	}
	catch (java.io.UnsupportedEncodingException ex)
	{
		throw new Exception(com.ibm.etools.cobol.importer.CobolResource.getString("_EXC_Invalid_codepage"));
	}
}

/**
 * @param text
 */
public static void validateCurrency(String text) throws Exception {	
	/* Validation rules:
	 *	1) The first character can be one of the four:  '  "  x  X
	 *	2) If starts with a ' or  " - the max length can be 3 and the end char can only be the one it was started with.
	 *	3) If starts with a X, the 2nd char and the 5th char should be " or '.
	 */
	boolean isValid = false;
	char[] textArray = new char[5];
    if ((text.startsWith("x") || text.startsWith("X") || text.startsWith("'") || text.startsWith("\""))){
    	text.getChars(0, text.length(), textArray, 0);
    	if (textArray[0] == '\'' || textArray[0] == '"') {
    		if (text.length() == 3) {
    			if (textArray[0] == textArray[2])
    				isValid = true;
    		}
    	} else if (textArray[0] == 'X' || textArray[0] == 'x'){
        	if (textArray[1] == '\'' || textArray[1] == '"') {    		
	    		if (text.length() == 5) {
	    			if (textArray[1] == textArray[4])
	    				isValid = true;
	    		}    		
        	}
    	}
	}    
    if (!(isValid))
		throw new Exception(com.ibm.etools.cobol.importer.CobolResource.getString("_EXC_Invalid_Currency"));
}

/**
 * This method, given any PICTURE string will expand it.
 */
public static String expandPictureString(String compressed) {
	int startIndex, endIndex, numberOfChars;
	String expandStr, expandChar;
	while (true) {
	  expandStr = "";
	  startIndex = compressed.indexOf("(");
	  if (startIndex == -1)
		break;
	  expandChar = compressed.substring(startIndex - 1, startIndex);
	  endIndex = compressed.indexOf(")");
	  numberOfChars = (new Integer(compressed.substring(startIndex + 1, endIndex))).intValue();
	  for (int i = 0; i < numberOfChars; i++)
		expandStr = expandStr + expandChar;
	  compressed = compressed.substring(0, startIndex - 1) + expandStr + compressed.substring(endIndex + 1);
	}
	return compressed;
}

/**
 * This method, given any PICTURE string will normalize it.
 * Ie. given 99V99 the normalized string would be 9(2)V9(2)
 * Note it is not necessarily the shortest pic representation.
 */
public static String normalizePictureString(String pic) {
	String out = "";
	char work[] = expandPictureString(pic).toCharArray();
	int inLen = work.length;
	int i, j;
	for (i = 0; i < inLen;) {
    if (i+1<inLen && work[i] == work [i+1]) {
		out = out + work[i] + "(";
		for (j = i; j+1 < inLen && work[j] == work[j+1];j++) {}
	    out = out + String.valueOf(j-i+1) +")";
		i = j+1;
	  } else {
		out = out + work[i];
		i++;
	  }
	}
	return out;
}
}
